# Line Bug program

LightSensor = A1
TapeThreshold = 0.45


def OnTape():
    return read(LightSensor) > TapeThreshold


RightWheel = M1
LeftWheel = M3
Speed = 1


def StartLeftTurn():
    talkto(LeftWheel)
    off()
    talkto(RightWheel)
    setright()
    setpower(Speed)
    on()


def StartRightTurn():
    talkto(RightWheel)
    off()
    talkto(LeftWheel)
    setleft()
    setpower(Speed)
    on()


def LineBug():
    while True:
        while OnTape():
            StartLeftTurn()

        while not OnTape():
            StartRightTurn()


LineBug()



#
# Copyright © 2019 Paulo Henrique Silva <ph.silva@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#
# Check assert success
#

assert 42 == 42
assert True
assert 1
assert [42]
assert "true"
assert "false"  # string s is True if len(s) > 0



#
# Copyright (c) 1980, 1993
#       The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the University nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

#
# snake - crt hack game.
#
# You move around the screen with arrow keys trying to pick up money
# without getting eaten by the snake.  hjkl work as in vi in place of
# arrow keys.  You can leave at the exit any time.
#

import time
import sys
import random
import curses
import math


def cashvalue():
    global loot
    global penalty
    return chunk * (loot - penalty) / 25


def _x(p):
    return p % 256 - 128


def _y(p):
    return p // 256 - 128


def _p(x, y):
    return (x + 128) + (y + 128) * 256


PENALTY = 10  # % penalty for invoking spacewarp

ME = "I"
SNAKEHEAD = "S"
SNAKETAIL = "s"
TREASURE = "$"
GOAL = "#"
LINES = 0
COLS = 0


def MIN(a, b):
    if a < b:
        return a
    return b


stdscr = 0


def erase():
    stdscr.erase()


def refresh():
    stdscr.refresh()


def getch():
    return stdscr.getch()


def move(x, y):
    stdscr.move(x, y)


def mvaddstr(y, x, str):
    stdscr.addstr(y, x, str)


def pchar(s, str):
    stdscr.addstr(_y(s) + 1, _x(s) + 1, str)
    refresh()


def endwin():
    curses.nocbreak()
    curses.echo()
    curses.endwin()


def delay(t):
    time.sleep(t * 0.05)


you = _p(0, 0)
money = _p(0, 0)
finish = _p(0, 0)
snake = [_p(0, 0)] * 6

loot = 0
penalty = 0
moves = 0
fast = 1

lcnt = 25
ccnt = 80

chunk = 0


def error(s):
    print(s)
    exit(1)


def main():
    global penalty, loot, finish, you, money, snake, chunk
    random.seed(time.monotonic())
    penalty = 0
    loot = 0
    initscr()
    i = MIN(lcnt, ccnt)
    if i < 4:
        error("screen too small for a fair game.")
    #
    #  chunk is the amount of money the user gets for each $.
    #  The formula below tries to be fair for various screen sizes.
    #  We only pay attention to the smaller of the 2 edges, since
    #  that seems to be the bottleneck.
    #  This formula is a hyperbola which includes the following points:
    #   (24, $25)       (original scoring algorithm)
    #   (12, $40)       (experimentally derived by the "feel")
    #   (48, $15)       (a guess)
    #  This will give a 4x4 screen $99/shot.  We don't allow anything
    #  smaller than 4x4 because there is a 3x3 game where you can win
    #  an infinite amount of money.
    #
    if i < 12:
        i = 12  # otherwise it isn't fair
    #
    #  Compensate for border.  This really changes the game since
    #  the screen is two squares smaller but we want the default
    #  to be $25, and the high scores on small screens were a bit
    #  much anyway.
    # /
    i += 2
    chunk = (675.0 / (i + 6)) + 2.5  #  min screen edge
    finish = snrand()
    you = snrand()
    money = snrand()
    snake[0] = snrand()
    for i in range(1, 6):
        if i < 6:
            snake[i] = chase(snake[i - 1])
    setup()
    mainloop()
    exit(0)


# Main command loop
def mainloop():
    global moves, money, loot, you
    repeat = 1
    lastc = 0
    while True:
        # Highlight you, not left & above
        move(_y(you) + 1, _x(you) + 1)
        refresh()
        c = getch()
        if c <= ord("9") and c >= ord("0"):
            repeat = c - ord("0")
            while True:
                c = getch()
                if c < ord("0") or ord("9") < c:
                    break
                repeat = 10 * repeat + ord(c) - ord("0")
        elif c != ord("."):
            repeat = 1
        if c == ord("."):
            c = lastc
        lastc = c
        if c == ord("x"):
            endwin()
            length(moves)
            exit(0)
        elif c == ord("p") or c == ord("d"):
            snap()
            continue
        elif c == ord("w"):
            spacewarp(0)
            continue
        elif c == ord("A"):
            repeat = _x(you)
            c = ord("h")
        elif c == ord("H") or c == ord("S"):
            repeat = _x(you) - _x(money)
            c = ord("h")
        elif c == ord("T"):
            repeat = _y(you)
            c = ord("k")
        elif c == ord("K") or c == ord("E"):
            repeat = _y(you) - _y(money)
            c = ord("k")
        elif c == ord("P"):
            repeat = ccnt - 1 - _y(you)
            c = ord("l")
        elif c == ord("L") or c == ord("F"):
            repeat = _x(money) - _x(you)
            c = ord("l")
        elif c == ord("B"):
            repeat = lcnt - 1 - _y(you)
            c = ord("j")
        elif c == ord("J") or c == ord("C"):
            repeat = _y(money) - _y(you)
            c = ord("j")
        for k in range(1, repeat + 1):
            moves += 1
            if c == ord("s") or c == ord("h") or c == ord("\b"):
                if _x(you) > 0:
                    pchar(you, " ")
                    you = _p(_x(you) - 1, _y(you))
                    pchar(you, ME)
            elif c == ord("f") or c == ord("l") or c == ord(" "):
                if _x(you) < ccnt - 1:
                    pchar(you, " ")
                    you = _p(_x(you) + 1, _y(you))
                    pchar(you, ME)
            elif c == ord("e") or c == ord("k") or c == ord("i"):
                if _y(you) > 0:
                    pchar(you, " ")
                    you = _p(_x(you), _y(you) - 1)
                    pchar(you, ME)
            elif c == ord("c") or c == ord("j") or c == ord("\n") or c == ord("m"):
                if _y(you) + 1 < lcnt:
                    pchar(you, " ")
                    you = _p(_x(you), _y(you) + 1)
                    pchar(you, ME)
            if you == money:
                loot += 25
                if k < repeat:
                    pchar(you, " ")
                while True:
                    money = snrand()
                    if money == finish:
                        continue
                    if money == you:
                        continue
                    if _x(money) < 5 and _y(money) == 0:
                        continue
                    break
                pchar(money, TREASURE)
                winnings(cashvalue())
                continue
            if you == finish:
                win(finish)
                endwin()
                print("You have won with $%d." % cashvalue())
                length(moves)
                exit(0)
            if pushsnake():
                print("pushsnake true")
                break


def initscr():
    global stdscr, lcnt, ccnt
    stdscr = curses.initscr()
    curses.noecho()
    curses.cbreak()
    COLS = 80
    LINES = 25
    lcnt = LINES - 2
    ccnt = COLS - 2
    erase()


#
# setup the board
#


def setup():
    pchar(you, ME)
    pchar(finish, GOAL)
    pchar(money, TREASURE)
    for i in range(1, 6):
        pchar(snake[i], SNAKETAIL)
    pchar(snake[0], SNAKEHEAD)
    drawbox()


def drawbox():
    for i in range(1, ccnt + 1):
        mvaddstr(0, i, "-")
        mvaddstr(lcnt + 1, i, "-")
    for i in range(0, lcnt + 2):
        mvaddstr(i, 0, "|")
        mvaddstr(i, ccnt + 1, "|")


def snrand():
    while True:
        p = _p(random.randrange(ccnt), random.randrange(lcnt))
        # make sure it's not on top of something else
        if _y(p) == 0 and _x(p) < 5:
            continue
        if p == you:
            continue
        if p == money:
            continue
        if p == finish:
            continue
        for i in range(6):
            if p == snake[i]:
                break
        else:
            break
    return p


dirs = ("N ", "NE", "E ", "SE", "S ", "SW", "W ", "NW")
mx = (0, 1, 1, 1, 0, -1, -1, -1)
my = (-1, -1, 0, 1, 1, 1, 0, -1)
absv = (1, 1.4, 1, 1.4, 1, 1.4, 1, 1.4)
oldw = 0


def chase(sp):
    global oldw, loot
    # this algorithm has bugs; otherwise the snake would get too good
    d = _p(_x(you) - _x(sp), _y(you) - _y(sp))
    v1 = math.sqrt(_x(d) ** 2 + _y(d) ** 2)
    w = 0
    max = 0
    wt = [0] * 8
    for i in range(8):
        vp = _x(d) * mx[i] + _y(d) * my[i]
        v2 = absv[i]
        if v1 > 0:
            vp = vp / (v1 * v2)
        else:
            vp = 1
        if vp > max:
            max = vp
            w = i
    for i in range(8):
        d = _p(_x(sp) + mx[i], _y(sp) + my[i])
        wt[i] = 0
        if _x(d) < 0 or _x(d) >= ccnt or _y(d) < 0 or _y(d) >= lcnt:
            continue
        #
        # Change to allow snake to eat you if you're on the money,
        # otherwise, you can just crouch there until the snake goes
        # away.  Not positive it's right.
        #
        # if d[1] == 0 and d[0] < 5: continue
        #
        if d == money:
            continue
        if d == finish:
            continue
        if i == w:
            wt[i] = loot // 10
        else:
            wt[i] = 1
        if i == oldw:
            wt[i] += loot // 20
    w = 0
    for i in range(8):
        w += wt[i]
    vp = random.randrange(w)
    vpo = vp
    for i in range(8):
        if vp < wt[i]:
            break
        else:
            vp -= wt[i]
    else:
        printw("failure\n")
        i = 0
    while wt[i] == 0:
        i += 1
    refresh()
    w = i
    oldw = w
    return _p(_x(sp) + mx[w], _y(sp) + my[w])


def spacewarp(w):
    global loot
    you = snrand()
    p = _p(COLS / 2 - 8, LINES / 2 - 1)
    if _x(p) < 0:
        p = _p(0, _y(p))
    if _y(p) < 0:
        p = _p(_x(p), 0)
    if w:
        str = "BONUS!!!"
        loot = loot - penalty
        penalty = 0
    else:
        str = "SPACE WARP!!!"
        penalty += loot / PENALTY
    for j in range(3):
        erase()
        refresh()
        delay(5)
        pchar(_p(_x(p) + 1, _y(p) + 1), str)
        refresh()
        delay(10)
    setup()
    winnings(cashvalue())


def snap():
    if not stretch(money):
        if not stretch(finish):
            pchar(you, "?")
            refresh()
            delay(10)
            pchar(you, ME)
    refresh()


def abs(x):
    if x < 0:
        return -x
    return x


def stretch(ps):
    p = you
    if abs(_x(ps) - _x(you)) < (ccnt / 12) and (_y(you) != _y(ps)):
        if _y(you) < _y(ps):
            for l in range(_y(you) + 1, _y(ps) + 1):
                p = _p(_x(p), l)
                pchar(p, "v")
            refresh()
            delay(10)
            while _y(p) > _y(you):
                chk(p)
                p = _p(_x(p), _y(p) - 1)
        else:
            p = _p(_x(p), _y(you) - 1)
            while _y(p) >= _y(ps):
                pchar(p, "^")
                p = _p(_x(p), _y(p) - 1)
            refresh()
            delay(10)
            while _y(p) < _y(you):
                chk(p)
                p = _p(_x(p), _y(p) + 1)
        return 1
    else:
        if abs(_y(ps) - _y(you)) < lcnt / 7 and _x(you) != _x(ps):
            p = _p(_x(p), _y(you))
            if _x(you) < _x(ps):
                p = _p(_x(you) + 1, _y(p))
                while _x(p) <= _x(ps):
                    pchar(p, ">")
                    p = _p(_x(p) + 1, _y(p))
                refresh()
                delay(10)
                while _x(p) > _x(you):
                    chk(p)
                    p = _p(_x(p) - 1, _y(p))
            else:
                p = _p(_x(you) - 1, _y(p))
                while _x(p) >= _x(ps):
                    pchar(p, "<")
                    p = _p(_x(p) - 1, _y(p))
                refresh()
                delay(10)
                while _x(p) < _x(you):
                    chk(p)
                    p = _p(_x(p) + 1, _y(p))
            return 1
    return 0


def surround(ps):
    if _x(ps) == 0:
        ps = _p(_x(ps) + 1, _y(ps))
    if _y(ps) == 0:
        ps = _p(_x(ps), _y(ps) + 1)
    if _y(ps) == LINES - 1:
        ps = _p(_x(ps), _y(ps) - 1)
    if _x(ps) == COLS - 1:
        ps = _p(_x(ps) - 1, _y(ps))
    mvaddstr(_y(ps), _x(ps), "/*\\")
    mvaddstr(_y(ps) + 1, _x(ps), "* *")
    mvaddstr(_y(ps) + 2, _x(ps), "\\*/")
    for j in range(20):
        pchar(ps, "@")
        refresh()
        delay(1)
        pchar(ps, " ")
        refresh()
        delay(1)
    mvaddstr(_y(ps), _x(ps), "   ")
    mvaddstr(_y(ps) + 1, _x(ps), "o.o")
    mvaddstr(_y(ps) + 2, _x(ps), "\\_/")
    refresh()
    delay(6)
    mvaddstr(_y(ps), _x(ps), "   ")
    mvaddstr(_y(ps) + 1, _x(ps), "o.-")
    mvaddstr(_y(ps) + 2, _x(ps), "\\_/")
    refresh()
    delay(6)
    mvaddstr(_y(ps), _x(ps), "   ")
    mvaddstr(_y(ps) + 1, _x(ps), "o.o")
    mvaddstr(_y(ps) + 2, _x(ps), "\\_/")
    refresh()
    delay(6)


def win(ps):
    boxsize = 10
    x = ps
    for j in range(1, boxsize):
        for k in range(j):
            pchar(x, "#")
            x = _p(_x(x), _y(x) - 1)
        for k in range(j):
            pchar(x, "#")
            x = _p(_x(x) + 1, _y(x))
        j += 1
        for k in range(j):
            pchar(x, "#")
            x = _p(_x(x), _y(x) + 1)
        for k in range(j):
            pchar(x, "#")
            x = _p(_x(x) - 1, _y(x))
        refresh()
        delay(1)


def pushsnake():
    global loot
    #
    # My manual says times doesn't return a value.  Furthermore, the
    # snake should get his turn every time no matter if the user is
    # on a fast terminal with typematic keys or not.
    # So I have taken the call to times out.
    #
    issame = 0
    for i in range(4, -1, -1):
        if snake[i] == snake[5]:
            issame += 1
    if issame == 0:
        pchar(snake[5], " ")
    # Need the following to catch you if you step on the snake's tail
    tmp = snake[5]
    for i in range(4, -1, -1):
        snake[i + 1] = snake[i]
    snake[0] = chase(snake[1])
    pchar(snake[1], SNAKETAIL)
    pchar(snake[0], SNAKEHEAD)
    for i in range(6):
        if snake[i] == you or tmp == you:
            surround(you)
            i = (cashvalue()) % 10
            bonus = random.randrange(10)
            mvaddstr(lcnt + 1, 0, "%d" % bonus)
            refresh()
            delay(30)
            if bonus == i:
                spacewarp(1)
                return 1
            endwin()
            if loot >= penalty:
                print("\nYou and your $%d have been eaten\n" % cashvalue())
            else:
                print("\nThe snake ate you.  You owe $%d.\n" % -cashvalue())
            length(moves)
            exit(0)
    return 0


def chk(sp):
    if sp == money:
        pchar(sp, TREASURE)
        return 2
    if sp == finish:
        pchar(sp, GOAL)
        return 3
    if sp == snake[0]:
        pchar(sp, SNAKEHEAD)
        return 4
    for j in range(1, 6):
        if sp == snake[j]:
            pchar(sp, SNAKETAIL)
            return 4
    if _x(sp) < 4 and _y(sp) == 0:
        winnings(cashvalue())
    if _y(you) == 0 and _x(you) < 4:
        pchar(you, ME)
        return 5
    if sp == you:
        pchar(sp, ME)
        return 1
    pchar(sp, " ")
    return 0


def winnings(won):
    if won > 0:
        pchar(_p(0, 0), "$%d" % won)


def length(num):
    print("You made %d moves.\n" % num)


main()



#
# Copyright © 2021 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#
# Test operator precedence
#

limit = 4

# test boolean operators

for a in (False, True):
    for b in (False, True):
        for c in (False, True):

            # not is higher than and

            assert (not a and b) == ((not a) and b)
            assert (b and not a) == (b and (not a))

            # and is higher than or

            assert (a or b and c) == (a or (b and c))
            assert (a and b or c) == ((a and b) or c)

# test boolean vs comparisons

for a in range(limit):
    for b in range(limit):
        for c in range(limit):
            for d in range(limit):

                # comparison is higher than boolean

                assert (a < b and c < d) == ((a < b) and (c < d))
                assert (a <= b and c <= d) == ((a <= b) and (c <= d))
                assert (a > b and c > d) == ((a > b) and (c > d))
                assert (a >= b and c >= d) == ((a >= b) and (c >= d))
                assert (a != b and c != d) == ((a != b) and (c != d))
                assert (a == b and c == d) == ((a == b) and (c == d))

# test bitwise vs comparisons

for a in range(limit):
    for b in range(limit):
        for c in range(limit):
            for d in range(limit):

                # bitwise is higher than comparison

                assert (a | b < c | d) == ((a | b) < (c | d))

# test bitwise operators

for a in range(limit):
    for b in range(limit):
        for c in range(limit):

            # & is higher than ^
            assert a ^ b & c == a ^ (b & c)
            assert a & b ^ c == (a & b) ^ c

            # ^ is higher than |
            assert a | b ^ c == a | (b ^ c)
            assert a ^ b | c == (a ^ b) | c



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
#
# Rotate a rainbow through a set of pixels
#

DOWN = BUTTONA
UP = BUTTONB

num_pixels = 10
bright = 0.05
pixels = [0] * num_pixels
for i in range(num_pixels):
    pixels[i] = [0, 0, 0]


def wheel(i, pos):
    global bright
    if pos < 0 or pos > 255:
        r = 0
        g = 0
        b = 0
    elif pos < 85:
        r = pos / 85
        g = 1 - r
        b = 0
    elif pos < 170:
        b = (pos - 85) / 85
        r = 1 - b
        g = 0
    else:
        g = (pos - 170) / 85
        r = 0
        b = 1 - g
    pixels[i][0] = r * bright
    pixels[i][1] = g * bright
    pixels[i][2] = b * bright


def rainbow_cycle(wait):
    global bright
    for j in range(255):
        for i in range(num_pixels):
            pixel_index = (i * 256 // num_pixels) + j
            wheel(i, pixel_index & 255)
        neopixel(pixels)
        if read(DOWN):
            if bright > 0.02:
                bright -= 0.01
                print(bright)
        elif read(UP):
            if bright < 0.99:
                bright += 0.01
                print(bright)
        time.sleep(wait)


def cycles():
    talkto(NEOPIXEL)
    while True:
        rainbow_cycle(0.001)


cycles()



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

import random
import curses
import time

stdscr = 0

snek = []
grow = 0
snak = 0
lines = 25
cols = 40
dx = 0
dy = 0

HEAD = "@"
TAIL = "+"
SNAK = "$"


def _x(p):
    return p % cols


def _y(p):
    return p // cols


def _p(x, y):
    return x + y * cols


def put_snak():
    global snek, snak
    while True:
        snak = _p(random.randrange(cols - 2) + 1, random.randrange(lines - 2) + 1)
        for s in snek:
            if snak == s:
                break
        else:
            break
    showp(snak, SNAK)


def showxy(x, y, str):
    stdscr.addstr(y, x, str)


def showp(pos, str):
    showxy(_x(pos), _y(pos), str)


miss = 0
hit_wall = 1
hit_snek = 2
hit_snak = 3


def move_snek():
    global snek, grow
    global dx, dy
    if dx or dy:
        old = snek[len(snek) - 1]
        sp = snek[0]
        nx = _x(sp) + dx
        ny = _y(sp) + dy
        if nx < 1 or nx >= cols - 1 or ny < 1 or ny >= lines - 1:
            return hit_wall
        new = _p(nx, ny)
        if grow > 0:
            tail = snek
            showxy(0, 0, "Score %d---" % (len(snek) + 1))
            grow -= 1
        else:
            tail = snek[:-1]
        for t in tail:
            if new == t:
                return hit_snek
        snek = [new] + tail
        showp(old, " ")
        if tail:
            showp(tail[0], TAIL)
        showp(new, HEAD)
        if new == snak:
            return hit_snak
    return miss


def done(msg):
    stdscr.move(lines - 1, 0)
    stdscr.refresh()
    stdscr.nodelay(False)
    curses.nocbreak()
    curses.echo()
    curses.endwin()
    print("You %s. Score %d" % (msg, len(snek)))
    exit(0)


def getch():
    return chr(stdscr.getch())


def main():
    global snek, dx, dy, snak, grow
    global stdscr, lines, cols
    stdscr = curses.initscr()
    curses.noecho()
    curses.cbreak()
    stdscr.nodelay(True)
    stdscr.erase()
    random.seed(time.monotonic())
    snek = [_p(1, 1)]
    put_snak()
    for x in range(1, cols - 1):
        showxy(x, 0, "-")
        showxy(x, lines - 1, "-")
    for y in range(1, lines - 1):
        showxy(0, y, "|")
        showxy(cols - 1, y, "|")
    showp(snek[0], HEAD)
    while True:
        stdscr.move(_y(snek[0]), _x(snek[0]))
        stdscr.refresh()
        time.sleep(0.1)
        c = getch()
        ndx = dx
        ndy = dy
        if c == "h":
            ndx = -1
            ndy = 0
        elif c == "j":
            ndx = 0
            ndy = 1
        elif c == "k":
            ndx = 0
            ndy = -1
        elif c == "l":
            ndx = 1
            ndy = 0
        elif c == "q" or c == "x":
            done("quit")
        elif c == "p":
            while getch() != "p":
                time.sleep(0.1)
        if ndx != -dx or ndy != -dy:
            dx = ndx
            dy = ndy
        hit = move_snek()
        if hit == hit_wall:
            done("hit the wall")
        elif hit == hit_snek:
            done("hit yourself")
        if hit == hit_snak:
            grow = 5
            put_snak()


main()



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

bright = 0.1


def rainbow(pos):
    if pos < 1:
        r = pos
        g = 1 - r
        b = 0
    elif pos < 2:
        b = pos - 1
        r = 1 - b
        g = 0
    else:
        g = pos - 2
        r = 0
        b = 1 - g
    neopixel(((bright * r, bright * g, bright * b),))


def blink():
    talkto(NEOPIXEL)
    while True:
        for pos in range(0, 3, 0.01):
            rainbow(pos)
            time.sleep(0.003)


blink()



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#
# Check short-circuiting and/or
#

or00 = (0 or 0) + 0
or10 = (1 or 0) + 0
or01 = (0 or 1) + 0
or11 = (1 or 1) + 0

and00 = (0 and 0) + 0
and10 = (1 and 0) + 0
and01 = (0 and 1) + 0
and11 = (1 and 1) + 0

if or00 != False:
    exit(1)
if or10 != True:
    exit(1)
if or01 != True:
    exit(1)
if or11 != True:
    exit(1)

if and00 != False:
    exit(1)
if and10 != False:
    exit(1)
if and01 != False:
    exit(1)
if and11 != True:
    exit(1)



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# Test scoping for op= expressions.
# In this case, there must be a local declaration of
# the variable before the operation, either a plain assignment
# or a global declaration
#
# a plain assignment will use the local variable
# a global declaration will use the global variable
#

i = 1


def fail(which):
    print("fail %s" % which)
    exit(1)


def global_decl():
    global i
    i += 1
    return i


if not global_decl() == 2:
    fail("global_decl() == 2")



#!/usr/bin/env python3
#
# Copyright © 2019–2020 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

import argparse
import curses
import serial
import serial.tools.list_ports
import sys

termios_error = serial.SerialException
if hasattr(serial, "termios"):
    termios_error = termios.error
import threading
import time

from curses import ascii
from pathlib import Path, PurePath

stdscr = 0

snek_lock = threading.Lock()

snek_current_window = 0
snek_edit_win = 0
snek_repl_win = 0

snek_cols = 0
snek_lines = 0

snek_monitor = False

snek_device = False

snek_dialog_waiting = False

snek_cur_path = None

snek_debug_file = False

baud_rates = [115200, 57600]

baud_selection = 0


def snek_debug(message):
    global snek_debug_file
    if not snek_debug_file:
        snek_debug_file = open("log", "w")
    snek_debug_file.write(message + "\n")
    snek_debug_file.flush()


# snek_debug("Starting snekde")

#
# Read a character from the keyboard, releasing the
# global lock while blocked so that serial input can arrive
#


def my_getch(edit_win):
    global snek_lock, snek_current_window, snek_dialog_waiting
    while True:
        edit_win.set_cursor()
        try:
            snek_lock.release()
            c = edit_win.window.getch()
        finally:
            snek_lock.acquire()
        if c == ord("\r"):
            c = ord("\n")
        if not snek_dialog_waiting:
            break
        if c == ord("\n"):
            snek_dialog_waiting.close()
            snek_dialog_waiting = False

    #
    # Check for resize
    #
    if c == curses.KEY_RESIZE:
        screen_resize()
    return c


class SnekPort:
    """Emulate new pySerial from old pySerial"""

    name = None
    description = None
    device = None

    def __init__(self, port):
        if type(port) is tuple:
            self.device = port[0]
            if len(port[1]) > 20:
                self.name = self.device
                if "/" in self.name:
                    self.name = self.name[self.name.rfind("/") + 1 :]
            else:
                self.name = port[1]
            self.description = port[1]
            self.hwid = port[2]
        else:
            self.device = port.device
            if port.name:
                self.name = port.name
            else:
                self.name = self.device
            self.description = port.description
            self.product = port.product
            self.hwid = port.hwid
        if self.name and "/" in self.name:
            self.name = self.name[self.name.rfind("/") + 1 :]


# A special hack for pre-3.5 pySerial
#
# That code is missing a call to 'str' to convert the check_output
# result to a string.

if "list_ports_posix" in serial.tools.__dict__:
    if hasattr(serial.tools.list_ports_posix, "popen"):
        del serial.tools.list_ports_posix.popen

        try:
            import subprocess
        except ImportError:

            def popen(argv):
                try:
                    si, so = os.popen4(" ".join(argv))
                    return so.read().strip()
                except:
                    raise IOError("lsusb failed")

        else:

            def popen(argv):
                try:
                    return str(
                        subprocess.check_output(argv, stderr=subprocess.STDOUT)
                    ).strip()
                except:
                    raise IOError("lsusb failed")

        serial.tools.list_ports_posix.popen = popen


def snek_list_ports():
    ports = []
    for port in serial.tools.list_ports.comports():
        ports += [SnekPort(port)]
    return ports


port_mods = {"239A:": "async"}


class SnekDevice:
    """Link to snek device"""

    serial = False
    reader_thread = False
    writer_thread = False
    alive = False
    _reader_alive = False
    interface = False
    write_queue = False
    device = ""
    interrupt_pending = False
    syncronous_put = False

    #
    # The interface needs to have a condition variable (cv) that is
    # signaled when data are available to write and function (receive)
    # that gets data that are read
    #

    def __init__(self, port, interface, rate=115200):
        self.interface = interface
        self.device = port.device

        # For devices without flow control, use snek's ENQ/ACK
        # mechanism to avoid overrunning the device input buffer

        self.synchronous_put = True
        self.synchronous_limit = 16

        for port_mod in port_mods:
            if port_mod in port.description or port_mod in port.hwid:
                if "async" in port_mods[port_mod]:
                    self.synchronous_put = False
        self.serial = serial.Serial(
            port=self.device,
            baudrate=rate,
            bytesize=serial.EIGHTBITS,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            xonxoff=False,
            rtscts=False,
            dsrdtr=False,
        )

    def start(self):
        """start worker threads"""

        self.alive = True
        self._reader_alive = True
        self.interrupt_pending = True

        # start threads

        self.receiver_thread = threading.Thread(target=self.reader, name="rx")
        self.receiver_thread.daemon = True
        self.receiver_thread.start()

        self.transmitter_thread = threading.Thread(target=self.writer, name="tx")
        self.transmitter_thread.daemon = True
        self.transmitter_thread.start()

        self.autobauder_thread = False
        self.autobauder_cancel = False
        if self.synchronous_put:
            self.autobauder_thread = threading.Thread(
                target=self.autobauder, name="baud"
            )
            self.autobauder_thread.daemon = True
            self.autobauder_thread.start()

    def stop_reader(self):
        if self.receiver_thread and threading.current_thread() != self.receiver_thread:
            self.serial.cancel_read()
            self.interface.cv.release()
            self.receiver_thread.join()
            self.interface.cv.acquire()

    def stop_writer(self):
        """set flag to stop worker threads"""
        if (
            self.transmitter_thread
            and threading.current_thread() != self.transmitter_thread
        ):
            self.interface.cv.notify()
            self.interface.cv.release()
            self.transmitter_thread.join()
            self.interface.cv.acquire()

    def stop_autobauder(self):
        if self.autobauder_thread:
            self.autobauder_cancel = True
            self.dc4.notify_all()
            self.autobauder_thread.join()

    def close(self):
        self.alive = False
        self.stop_reader()
        self.stop_writer()
        self.stop_autobauder()
        try:
            self.serial.write_timeout = 1
            self.serial.write(b"\x0f")
            self.serial.reset_output_buffer()
            self.serial.close()
        except serial.SerialException:
            pass

    def reader(self):
        """loop and copy serial->"""
        try:
            while self.alive and self._reader_alive:
                # read all that is there or wait for one byte
                data = self.serial.read(1)
                if data:
                    self.interface.receive(data)
        except (serial.SerialException, termios_error):
            # snek_debug('serial interface failed in reader')
            self.interface.failed(self.device)
        finally:
            self.receiver_thread = False

    def writer(self):
        global snek_lock

        """Copy queued data to the serial port."""
        try:
            sent = 0
            while True:
                send_data = b""
                interrupt = False
                # snek_debug('writer locking')
                with snek_lock:
                    # snek_debug('writer waiting')

                    # Wait until there's something to do

                    while (
                        not self.write_queue
                        and not self.interrupt_pending
                        and self.alive
                    ):
                        self.interface.cv.wait()
                    # snek_debug('writer awake')
                    send_data = self.write_queue
                    interrupt = self.interrupt_pending
                    self.write_queue = False
                    self.interrupt_pending = False
                if not self.alive:
                    break
                if interrupt:
                    # snek_debug('interrupt')
                    self.serial.reset_output_buffer()
                    self.serial.write(b"\x0f\x03\x0e")
                if self.synchronous_put:
                    while send_data:
                        amt = self.synchronous_limit - sent
                        if amt > len(send_data):
                            amt = len(send_data)
                        this_time = send_data[:amt]
                        send_data = send_data[amt:]
                        self.serial.write(this_time)
                        sent += amt
                        if sent >= self.synchronous_limit:
                            # snek_debug('sync %d limit %d' % (sent, self.synchronous_limit))
                            snek_monitor.sync()
                            sent = 0
                else:
                    if send_data:
                        self.serial.write(send_data)
        except (serial.SerialException, termios_error):
            # snek_debug('serial interface failed in writer')
            self.interface.failed(self.device)
        finally:
            self.transmitter_thread = False

    def do_autobaud(self, delay=0):
        global baud_selection, snek_lock
        with snek_lock:
            if delay > 0:
                time.sleep(delay)
            old_baud = baud_selection
            for i in range(len(baud_rates)):
                baud_selection = i
                # snek_debug("try baud %d" % baud_rates[i])
                if self.autobauder_cancel:
                    break
                self.set_baud(baud_rates[baud_selection])
                if snek_monitor.ping():
                    return True
            baud_selection = old_baud
            self.set_baud(baud_rates[baud_selection])
            return False

    def autobauder(self):
        global baud_selection

        # Try to autobaud right away, then wait 1.5 seconds
        # for the device to boot and try again
        if not self.do_autobaud(0.1):
            self.do_autobaud(1.5)

        self.autobauder_thread = False

    def interrupt(self):
        # snek_debug('pend interrupt waited')
        self.interrupt_pending = True
        self.interface.cv.notify()
        # snek_debug('pended')

    def writeb(self, data):
        # snek_debug('writeb waited')
        if self.write_queue:
            self.write_queue += data
        else:
            self.write_queue = data
        self.interface.cv.notify()

    def write(self, data):
        self.writeb(data.encode("utf-8"))

    def command(self, data, intr="\x03"):
        self.write("\x0e" + intr + data)

    def set_baud(self, baud):
        self.serial.baudrate = baud


class EditWin:
    """Editable text object"""

    window = 0
    lines = 0
    cols = 0
    y = 0
    point = 0
    top_line = 0
    tab_width = 4
    text = ""
    cut = ""
    mark = -1
    changed = False

    undo = []

    def __init__(self, lines, cols, y, x):
        self.lines = lines
        self.cols = cols
        self.y = y
        self.window = curses.newwin(lines, cols, y, x)
        self.window.keypad(True)
        self.window.notimeout(False)

    # Set contents, resetting state back to start

    def set_text(self, text):
        self.text = text
        self.point = 0
        self.mark = -1
        self.top_line = 0
        self.changed = False

    # Convert text index to x/y coord

    def point_to_cursor(self, point):
        line = -1
        col = 0
        for s in self.text[:point].split("\n"):
            line += 1
            col = len(s)
        return (col, line)

    # Convert x/y coord to text index

    def cursor_to_point(self, cursor):
        (cur_col, cur_line) = cursor
        if cur_line < 0:
            cur_line = 0
        elif cur_line >= len(self.text.split("\n")):
            cur_line = len(self.text.split("\n")) - 1
        bol = 0
        line = 0
        point = 0
        for s in self.text.split("\n"):
            point = bol + min(cursor[0], len(s))
            if line == cur_line:
                break
            line += 1
            bol += len(s) + 1
        return point

    # Make sure current point is visible in the window

    def scroll_to_point(self):
        while self.point_to_cursor(self.point)[1] < self.top_line:
            self.top_line -= 1
        while self.point_to_cursor(self.point)[1] >= self.top_line + self.lines:
            self.top_line += 1

    def addstr(self, line, col, text, attrib=None):
        if col < self.cols:
            if line == self.lines - 1:
                text = text[: self.cols - col - 1]
            else:
                text = text[: self.cols - col]
            if attrib:
                self.window.addstr(line, col, text, attrib)
            else:
                self.window.addstr(line, col, text)

    # Repaint the window

    def repaint(self):
        self.window.erase()
        self.scroll_to_point()
        selection = self.get_selection()
        if selection:
            start_selection = self.point_to_cursor(selection[0])
            end_selection = self.point_to_cursor(selection[1])
        line = 0
        for s in self.text.split("\n"):
            if self.top_line <= line and line < self.top_line + self.lines:

                # Paint the marked region in reverse video

                if (
                    selection
                    and start_selection[1] <= line
                    and line <= end_selection[1]
                ):
                    if line == start_selection[1]:
                        before = s[: start_selection[0]]
                        middle = s[start_selection[0] :]
                    else:
                        before = ""
                        middle = s
                    if line == end_selection[1]:
                        after = middle[end_selection[0] :]
                        middle = middle[: end_selection[0]]
                    else:
                        after = ""
                    self.addstr(line - self.top_line, 0, before)
                    self.addstr(
                        line - self.top_line, len(before), middle, curses.A_REVERSE
                    )
                    self.addstr(line - self.top_line, len(before) + len(middle), after)
                else:
                    self.addstr(line - self.top_line, 0, s)
            line += 1
        self.window.refresh()

    # Set window size

    def resize(self, lines, cols, y, x):
        self.lines = lines
        self.cols = cols
        self.window.resize(lines, cols)
        self.window.mvwin(y, x)
        self.repaint()

    # This window is the input window, set the cursor position
    # to the current point

    def set_cursor(self):
        p = self.point_to_cursor(self.point)
        col = min(self.cols - 1, p[0])
        self.window.move(p[1] - self.top_line, col)
        self.window.refresh()

    # Find the indent of the specified line

    def indent_at(self, line):
        bol = self.cursor_to_point((0, line))
        point = bol
        while point < len(self.text) and self.text[point] == " ":
            point += 1
        return point - bol

    # Return the last character on the specified line

    def last_ch(self, line):
        eol = self.cursor_to_point((65536, line))
        point = eol - 1
        while point > 0 and self.text[point] == " ":
            point -= 1
        return self.text[point]

    # Is 'point' in the indentation of its line?

    def in_indent(self, point):
        cursor = self.point_to_cursor(point)
        if cursor[0] > 0 and self.indent_at(cursor[1]) >= cursor[0]:
            return True
        return False

    # Move left

    def left(self):
        if self.point > 0:
            self.point -= 1

    # Move up

    def up(self):
        pos = self.point_to_cursor(self.point)
        point = self.cursor_to_point((pos[0], pos[1] - 1))
        if point == self.point:
            point = self.cursor_to_point((0, pos[1]))
        self.point = point

    # Move right

    def right(self):
        if self.point < len(self.text):
            self.point += 1

    # Move down

    def down(self):
        pos = self.point_to_cursor(self.point)
        point = self.cursor_to_point((pos[0], pos[1] + 1))
        if point == self.point:
            point = self.cursor_to_point((65536, pos[1]))
        self.point = point

    # Move to beginning of line

    def bol(self):
        pos = self.point_to_cursor(self.point)
        self.point = self.cursor_to_point((0, pos[1]))

    # Move to end of line

    def eol(self):
        pos = self.point_to_cursor(self.point)
        self.point = self.cursor_to_point((65536, pos[1]))

    # Move to beginning of buffer

    def bob(self):
        self.point = 0

    # Move to end of buffer

    def eob(self):
        self.point = len(self.text)

    # Move to previous page. That means moving up by self.lines-1 and
    # then placing that line at the bottom of the screen

    def prev_page(self):
        pos = self.point_to_cursor(self.point)
        self.point = self.cursor_to_point((pos[0], pos[1] - (self.lines) - 1))
        new_line = self.point_to_cursor(self.point)[1]
        top_line = new_line - (self.lines - 1)
        top_pos = self.cursor_to_point((0, top_line))
        self.top_line = self.point_to_cursor(top_pos)[1]

    # Move to next page. This means moving down by self.lines-1 and
    # then placing that line at the top of the screen

    def next_page(self):
        pos = self.point_to_cursor(self.point)
        self.point = self.cursor_to_point((pos[0], pos[1] + (self.lines) - 1))
        self.top_line = self.point_to_cursor(self.point)[1]

    # Check whether in the last line of the buffer

    def in_last_line(self):
        end = self.point_to_cursor(len(self.text))
        return self.point >= self.cursor_to_point((0, end[1]))

    def push_undo(self, point, operation):
        self.undo.append((point, operation, self.point, self.mark))

    def pop_undo(self):
        if not self.undo:
            return False
        (point, operation, self_point, self_mark) = self.undo.pop()

        if isinstance(operation, str):
            # Replace deleted text
            self.text = self.text[:point] + operation + self.text[point:]
        else:
            # Delete inserted text
            self.text = self.text[:point] + self.text[point + operation :]

        self.point = self_point
        self.mark = self_mark
        self.changed = True
        return True

    # Insert some text, adjusting self.point and self.mark if the text
    # is before them

    def insert(self, point, text):
        self.push_undo(point, len(text))
        self.text = self.text[:point] + text + self.text[point:]
        if point < self.point:
            self.point += len(text)
        if point < self.mark:
            self.mark += len(text)
        self.changed = True

    def insert_at_point(self, text):
        self.insert(self.point, text)
        self.point += len(text)

    # Delete some text, adjusting self.point if the delete starts
    # before it

    def _adjust_delete_position(
        self, delete_point, delete_count, moving_point, is_mark
    ):
        if delete_point <= moving_point and moving_point < delete_point + delete_count:
            if is_mark:
                return -1
            moving_point = delete_point
        elif delete_point + delete_count <= moving_point:
            moving_point -= delete_count
        return moving_point

    def delete(self, point, count):
        self.push_undo(point, self.text[point : point + count])
        self.text = self.text[:point] + self.text[point + count :]
        self.point = self._adjust_delete_position(point, count, self.point, False)
        if self.mark >= 0:
            self.mark = self._adjust_delete_position(point, count, self.mark, True)
        self.changed = True

    def delete_at_point(self, count):
        self.delete(self.point, count)

    # Delete back to the previous tab stop

    def backtab(self):
        pos = self.point_to_cursor(self.point)
        if pos[0] == 0:
            return
        to_remove = pos[0] % self.tab_width
        if to_remove == 0:
            to_remove = self.tab_width
        self.point -= to_remove
        self.delete_at_point(to_remove)

    # Delete something. If there's a mark, delete that.  otherwise,
    # delete backwards, if in indent of the line, backtab

    def backspace(self):
        selection = self.get_selection()
        if selection:
            self.delete(selection[0], selection[1] - selection[0])
            self.mark = -1
        elif self.point > 0:
            if self.in_indent(self.point):
                self.backtab()
            else:
                self.left()
                self.delete_at_point(1)

    # Set or clear the 'mark', which defines
    # the other end of the current selection

    def toggle_mark(self):
        if self.mark >= 0:
            self.mark = -1
        else:
            self.mark = self.point

    # Return the extent of the current selection, False if none

    def get_selection(self):
        if self.mark >= 0:
            return (min(self.mark, self.point), max(self.mark, self.point))
        else:
            return False

    # Copy from mark to point and place in cut buffer
    # delete from text if requested

    def copy(self, delete=False):
        selection = self.get_selection()
        if selection:
            (start, end) = selection

            self.cut = self.text[start:end]
            if delete:
                self.delete(start, end - start)
            self.mark = -1

    # Paste any cut buffer at point

    def paste(self):
        if self.cut:
            self.insert_at_point(self.cut)

    # Set indent of current line to 'want'. Leave point
    # at the end of the indent

    def indent(self, want):
        self.bol()
        have = 0
        while self.point < len(self.text) and self.text[self.point] == " ":
            self.right()
            have += 1
        if have < want:
            self.insert_at_point(" " * (want - have))
        elif have > want:
            self.delete_at_point(have - want)

    # Automatically indent the current line,
    # using the previous line as a guide

    def auto_indent(self):
        cursor = self.point_to_cursor(self.point)
        want = 0
        if cursor[1] > 0:
            want = self.indent_at(cursor[1] - 1)
            if self.last_ch(cursor[1] - 1) == ":":
                want += self.tab_width
        self.indent(want)

    # Delete to end of line, or delete newline if at end of line

    def delete_to_eol(self):
        current = self.point_to_cursor(self.point)
        eol = self.cursor_to_point((65536, current[1]))
        if self.point == eol:
            self.delete_at_point(1)
        else:
            self.delete_at_point(eol - self.point)

    # Read a character for this window

    def getch(self):
        self.repaint()
        return my_getch(self)

    # Return the contents of the previous line

    def cur_line(self):
        pos = self.point_to_cursor(self.point)
        start = self.cursor_to_point((0, pos[1]))
        end = self.cursor_to_point((65536, pos[1]))
        return self.text[start:end]

    def dispatch(self, ch):
        if ch == 0 or ch == ord(" ") + 0x80:
            self.toggle_mark()
        elif ch == ord("c") & 0x1F or ch == ord("c") + 0x80:
            self.copy(delete=False)
        elif ch == ord("x") & 0x1F or ch == ord("x") + 0x80:
            self.copy(delete=True)
        elif ch == ord("v") & 0x1F or ch == ord("v") + 0x80:
            self.paste()
        elif ch == ord("k") & 0x1F:
            self.delete_to_eol()
        elif ch == ord("z") & 0x1F:
            self.pop_undo()
        elif ch == curses.KEY_LEFT or ch == ord("b") & 0x1F:
            self.left()
        elif ch == curses.KEY_RIGHT or ch == ord("f") & 0x1F:
            self.right()
        elif ch == curses.KEY_UP or ch == ord("p") & 0x1F:
            self.up()
        elif ch == curses.KEY_DOWN or ch == ord("n") & 0x1F:
            self.down()
        elif ch == curses.KEY_HOME or ch == ord("a") & 0x1F:
            self.bol()
        elif ch == curses.KEY_PPAGE:
            self.prev_page()
        elif ch == curses.KEY_NPAGE:
            self.next_page()
        elif ch == curses.KEY_END or ch == ord("e") & 0x1F:
            self.eol()
        elif ch == ord("\t"):
            self.auto_indent()
        elif ch in (curses.ascii.BS, curses.KEY_BACKSPACE, curses.ascii.DEL):
            self.backspace()
        elif curses.ascii.isprint(ch) or ch == ord("\n"):
            self.insert_at_point(chr(ch))


class ErrorWin:
    """Show an error message"""

    label = ""
    x = 0
    y = 0
    nlines = 5
    ncols = 40
    inputthread = True

    window = False

    def __init__(self, label, inputthread=True):
        self.label = label
        self.inputthread = inputthread
        self.nlines = 5
        self.ncols = min(snek_cols, max(40, len(label) + 2))
        self.x = (snek_cols - self.ncols) // 2
        self.y = (snek_lines - self.nlines) // 2
        self.window = curses.newwin(self.nlines, self.ncols, self.y, self.x)
        self.window.keypad(True)
        self.window.notimeout(False)
        self.run_dialog()

    def repaint(self):
        self.window.border()
        l = len(self.label)
        if l > self.ncols:
            l = self.ncols
        self.window.addstr(1, (self.ncols - l) // 2, self.label)
        self.window.addstr(3, 2, "OK")

    def close(self):
        del self.window
        screen_repaint()

    def run_dialog(self):
        global snek_dialog_waiting
        self.repaint()
        self.window.move(3, 4)
        self.window.refresh()
        if self.inputthread:
            self.window.getstr()
            self.close()
        else:
            snek_dialog_waiting = self


class GetTextWin:
    """Prompt for line of text"""

    label = ""
    prompt = ""
    x = 0
    y = 0
    nlines = 5
    ncols = 40

    window = False

    def __init__(self, label, prompt="File:"):
        self.label = label
        self.prompt = prompt
        self.x = (snek_cols - self.ncols) // 2
        self.y = (snek_lines - self.nlines) // 2
        self.window = curses.newwin(self.nlines, self.ncols, self.y, self.x)
        self.window.keypad(True)
        self.window.notimeout(False)

    def repaint(self):
        self.window.border()
        self.window.addstr(1, (self.ncols - len(self.label)) // 2, self.label)
        self.window.addstr(3, 2, self.prompt)

    def run_dialog(self):
        self.repaint()
        self.window.move(3, 8)
        curses.echo()
        name = self.window.getstr()
        curses.noecho()
        del self.window
        screen_repaint()
        return str(name, encoding="utf-8", errors="ignore")


class GetFileWin:
    """Prompt for a filename"""

    label = ""
    dir = ""
    x = 0
    y = 0
    nlines = 20
    ncols = 60
    new = False
    window = False
    cur_ent = 1
    file = ""
    full_file = ""

    def __init__(self, label, path, new=False):
        self.label = label
        self.new = new
        self.nlines = snek_lines - 2
        self.ncols = snek_cols - 10
        self.cur_ent = 1
        self.top_ent = 0
        self.x = (snek_cols - self.ncols) // 2
        self.y = (snek_lines - self.nlines) // 2
        self.window = curses.newwin(self.nlines, self.ncols, self.y, self.x)
        self.window.keypad(True)
        self.window.notimeout(False)
        if path:
            self.dir = path.parent
        else:
            self.dir = Path().resolve()
        self.get_files()
        if path:
            for i in range(len(self.files)):
                if self.files[i] == path.name:
                    self.set_ent(i)
                    break

    def get_files(self):
        files = []
        dirs = []
        for f in self.dir.glob("*"):
            if f.is_dir():
                dirs += [f.name]
            else:
                files += [f.name]
        dirs.sort()
        files.sort()
        self.files = [".."] + dirs + files
        self.file_offset = 0
        self.cur_ent = 1
        self.file = self.files[1]

    def set_dir(self, dir):
        self.dir = self.path.resolve()
        self.get_files()

    def ent_line(self, i):
        return 4 + i

    def ents(self):
        return self.nlines - 5

    def set_ent(self, ent):
        if 0 <= ent and ent < len(self.files):
            self.cur_ent = ent
            self.file = self.files[self.cur_ent]
            if ent < self.file_offset:
                self.file_offset = ent
            elif ent >= self.file_offset + self.ents():
                self.file_offset = ent - self.ents() + 1

    def repaint(self):
        self.window.border()
        self.window.addstr(1, (self.ncols - len(self.label)) // 2, self.label)
        self.path = self.dir / PurePath(self.file)
        if self.file == "":
            self.full_file = str(self.dir / PurePath("x"))[:-1]
        else:
            self.full_file = str(self.path)
        self.window.addstr(2, 2, " " * (self.ncols - 3))
        left = max(0, len(self.full_file) - 2 - self.ncols)
        self.window.addstr(2, 2, self.full_file[left:])
        for i in range(0, self.ents()):
            f = self.file_offset + i
            self.window.addstr(self.ent_line(i), 2, " " * (self.ncols - 3))
            if f < len(self.files):
                text = self.files[f]
                if (self.dir / PurePath(text)).is_dir():
                    text = "[" + text + "]"
                text = text[0 : self.ncols - 3]
                if f == self.cur_ent:
                    self.window.addstr(self.ent_line(i), 2, text, curses.A_REVERSE)
                else:
                    self.window.addstr(self.ent_line(i), 2, text)

    def set_cursor(self):
        self.window.move(2, min(2 + len(self.full_file), self.ncols - 2))
        self.window.refresh()

    def run_dialog(self):
        while True:
            self.repaint()
            self.window.refresh()
            ch = my_getch(self)
            if ch == ord("\n"):
                if not self.path.is_dir():
                    break
                self.set_dir(self.path.resolve())
            elif ch == curses.ascii.ESC or ch == curses.KEY_F4:
                self.path = None
                break
            elif ch == curses.KEY_UP or ch == ord("p") & 0x1F:
                self.set_ent(self.cur_ent - 1)
            elif ch == curses.KEY_DOWN or ch == ord("n") & 0x1F:
                self.set_ent(self.cur_ent + 1)
            else:
                if not self.new:
                    curses.beep()
                elif ch in (curses.ascii.BS, curses.KEY_BACKSPACE, curses.ascii.DEL):
                    if len(self.file) > 0:
                        self.file = self.file[0:-1]
                elif curses.ascii.isprint(ch):
                    self.file = self.file + chr(ch)
                else:
                    curses.beep()

        del self.window
        screen_repaint()
        return self.path


class GetPortWin:
    """Prompt for serial port"""

    ports = False
    x = 0
    y = 0
    nlines = 0
    ncols = 40
    curline = 1
    fakeport = None
    title = None

    def label(self, port):
        l = "%-15.15s %s" % (port.name, port.description)
        return l[: snek_cols - 3]

    def make_fake(self, description):
        return SnekPort(("Cancel", "Cancel", description))

    def __init__(self, title="Select Port"):
        self.title = title
        self.ports = snek_list_ports()
        if len(self.ports) == 0:
            self.fakeport = self.make_fake("No ports available")
        else:
            self.fakeport = self.make_fake("")
        maxlen = len(self.title)
        for port in self.ports:
            maxlen = max(maxlen, len(self.label(port)))
        maxlen = max(maxlen, len(self.label(self.fakeport)))
        self.ncols = maxlen + 3
        self.nlines = 2 + 1 + len(self.ports) + 1
        self.curline = 0
        self.x = (snek_cols - self.ncols) // 2
        self.y = (snek_lines - self.nlines) // 2

        self.window = curses.newwin(self.nlines, self.ncols, self.y, self.x)
        self.window.keypad(True)
        self.window.notimeout(False)

    def paint_port(self, line, port):
        if line == self.curline:
            self.window.addstr(line + 2, 1, " " * (self.ncols - 3), curses.A_REVERSE)
            self.window.addstr(line + 2, 1, self.label(port), curses.A_REVERSE)
        else:
            self.window.addstr(line + 2, 1, " " * (self.ncols - 3))
            self.window.addstr(line + 2, 1, self.label(port))

    def repaint(self):
        self.window.border()
        self.window.addstr(1, (self.ncols - len(self.title)) // 2, self.title)
        line = 0
        for port in self.ports:
            self.paint_port(line, port)
            line += 1
        self.paint_port(line, self.fakeport)

    def set_cursor(self):
        self.window.move(self.curline + 2, self.ncols - 2)

    def run_dialog(self):
        while True:
            self.repaint()
            ch = my_getch(self)
            if ch == ord("\n"):
                del self.window
                screen_repaint()
                if self.curline >= len(self.ports):
                    return None
                return self.ports[self.curline]
            elif ch == curses.KEY_UP or ch == ord("p") & 0x1F:
                self.curline = max(0, self.curline - 1)
            elif ch == curses.KEY_DOWN or ch == ord("n") & 0x1F:
                self.curline = min(len(self.ports), self.curline + 1)


def screen_get_sizes():
    global snek_lines, snek_cols, stdscr
    snek_lines, snek_cols = stdscr.getmaxyx()
    repl_lines = snek_lines // 3
    edit_lines = snek_lines - repl_lines - 2
    edit_y = 1
    repl_y = edit_y + edit_lines + 1
    return (edit_lines, edit_y, repl_lines, repl_y)


help_text = (
    ("F1", "Device"),
    ("F2", "Get"),
    ("F3", "Put"),
    ("F4", "Quit"),
    ("F5", "Load"),
    ("F6", "Save"),
    ("F7", "Switch"),
    ("F8", "Rate"),
)

# Paint the function key help text and the separator line


def screen_paint():
    global stdscr, snek_device, snek_edit_win
    help_col = 0
    help_cols = min(snek_cols // len(help_text), 13)
    stdscr.addstr(0, 0, " " * snek_cols)
    for (f, t) in help_text:
        stdscr.addstr(0, help_col, " %2s: %-6s " % (f, t), curses.A_REVERSE)
        help_col += help_cols
    device_name = "<no device>"
    if snek_device:
        device_name = snek_device.device
    device_name = "%s %d" % (device_name, baud_rates[baud_selection])
    device_col = snek_cols - len(device_name)
    if device_col < 0:
        device_col = 0
    mid_y = snek_edit_win.y + snek_edit_win.lines
    stdscr.addstr(mid_y, device_col, device_name, curses.A_REVERSE)
    if device_col >= 6:
        stdscr.addstr(mid_y, device_col - 6, "      ", curses.A_REVERSE)
    for col in range(0, device_col - 6, 5):
        stdscr.addstr(mid_y, col, "snek ", curses.A_REVERSE)
    stdscr.refresh()


# Repaint everything, as when a dialog goes away


def screen_repaint():
    global snek_edit_win, snek_repl_win, snek_current_window
    snek_edit_win.repaint()
    snek_repl_win.repaint()
    screen_paint()
    if snek_current_window:
        snek_current_window.set_cursor()


# Handle screen resize


def screen_resize():
    global snek_edit_win, snek_repl_win
    #
    # Update desired window sizes
    #
    (edit_lines, edit_y, repl_lines, repl_y) = screen_get_sizes()
    stdscr.clear()
    snek_edit_win.resize(edit_lines, snek_cols, edit_y, 0)
    snek_repl_win.resize(repl_lines, snek_cols, repl_y, 0)
    screen_paint()


def screen_init(text):
    global stdscr, snek_edit_win, snek_repl_win
    stdscr = curses.initscr()
    curses.noecho()
    curses.raw()
    stdscr.keypad(True)
    stdscr.notimeout(False)
    (edit_lines, edit_y, repl_lines, repl_y) = screen_get_sizes()
    snek_edit_win = EditWin(edit_lines, snek_cols, edit_y, 0)
    if text:
        snek_edit_win.set_text(text)
    snek_repl_win = EditWin(repl_lines, snek_cols, repl_y, 0)
    screen_paint()


def screen_fini():
    global stdscr
    stdscr.keypad(False)
    curses.noraw()
    curses.echo()
    curses.endwin()


def switch_baud():
    global snek_device, baud_selection, baud_rates
    baud_selection = (baud_selection + 1) % len(baud_rates)
    if snek_device:
        snek_device.set_baud(baud_rates[baud_selection])
    screen_paint()


def snekde_open_device():
    global snek_device, snek_monitor, baud_selection, baud_rates
    dialog = GetPortWin()
    port = dialog.run_dialog()
    if not port:
        return
    try:
        old_device = snek_device
        snek_device = SnekDevice(port, snek_monitor, rate=baud_rates[baud_selection])
        snek_device.start()
        if old_device:
            old_device.close()
            del old_device

        screen_paint()
    except OSError as e:
        message = e.strerror
        if not message:
            message = "failed"
        ErrorWin("%s: %s" % (port.name, message))


def snekde_get_text():
    global snek_edit_win, snek_device
    snek_edit_win.set_text("")
    snek_device.command("eeprom.show(1)\n")


def snekde_put_text():
    global snek_edit_win, snek_device, snek_monitor
    if len(snek_edit_win.text.strip()) == 0:
        ErrorWin("No program to put")
        return
    snek_device.command("eeprom.write()\n")
    text = snek_edit_win.text.encode("utf-8")
    snek_device.writeb(text)
    snek_device.writeb(b"\004")

    snek_device.command("reset()\n", intr="")
    snek_edit_win.changed = False


#
# Load a path into the edit buffer
#
def load_file(path):
    global snek_cur_path
    with open(path, "r") as myfile:
        data = myfile.read()
        snek_edit_win.set_text(data)
        snek_edit_win.repaint()
        snek_cur_path = path
        snek_edit_win.changed = False


#
# Save the edit buffer to a path
#
def save_file(path):
    global snek_cur_path
    with open(path, "w") as myfile:
        myfile.write(snek_edit_win.text)
        snek_cur_path = path
        snek_edit_win.changed = False


#
# Display a dialog box requesting a file name,
# then load that file
#
def snekde_load_file():
    global snek_edit_win, snek_cur_path
    dialog = GetFileWin("Load File", snek_cur_path, new=False)
    path = dialog.run_dialog()
    if not path:
        return
    try:
        load_file(path)
    except OSError as e:
        ErrorWin("%s: %s" % (e.filename, e.strerror))


#
# Display a dialog box requesting a file name,
# then save to that file
#
def snekde_save_file():
    global snek_edit_win, snek_cur_path
    dialog = GetFileWin("Save File", snek_cur_path, new=True)
    path = dialog.run_dialog()
    if not path:
        return
    if path != snek_cur_path and path.exists():
        ErrorWin("%s: already exists" % str(path))
    try:
        save_file(path)
    except OSError as e:
        ErrorWin("%s: %s" % (e.filename, e.strerror))


def run():
    global snek_current_window, snek_edit_win, snek_repl_win, snek_device, snek_lock
    snek_current_window = snek_edit_win
    prev_exit = False
    prev_get = False
    snek_lock.acquire()
    while True:
        ch = snek_current_window.getch()
        if ch == curses.ascii.ESC:
            ch = snek_current_window.getch() | 0x80
        if ch == 3:
            if snek_device:
                snek_device.interrupt()
        elif ch == curses.KEY_F1 or ch == ord("1") | 0x80:
            snekde_open_device()
        elif ch == curses.KEY_F2 or ch == ord("2") | 0x80:
            if snek_device:
                if snek_edit_win.changed and not prev_get:
                    ErrorWin("Unsaved changes, get again to abandon them")
                    prev_get = True
                    continue
                snekde_get_text()
            else:
                ErrorWin("No device")
        elif ch == curses.KEY_F3 or ch == ord("3") | 0x80:
            if snek_device:
                snekde_put_text()
            else:
                ErrorWin("No device")
        elif ch == curses.KEY_F4 or ch == ord("4") | 0x80:
            if snek_edit_win.changed and not prev_exit:
                ErrorWin("Unsaved changes, quit again to abandon them")
                prev_exit = True
                continue
            sys.exit(0)
        elif ch == curses.KEY_F5 or ch == ord("5") | 0x80:
            snekde_load_file()
        elif ch == curses.KEY_F6 or ch == ord("6") | 0x80:
            snekde_save_file()
        elif ch == curses.KEY_F7 or ch == ord("7") | 0x80 or ch == ord("o") & 0x1F:
            if snek_current_window is snek_edit_win:
                snek_current_window = snek_repl_win
            else:
                snek_current_window = snek_edit_win
        elif ch == curses.KEY_F8 or ch == ord("8") | 0x80:
            switch_baud()
        elif ch == ord("\n"):
            if snek_current_window is snek_edit_win:
                snek_current_window.dispatch(ch)
                snek_current_window.auto_indent()
            elif snek_device:
                data = snek_repl_win.cur_line()
                #
                # Trim off snek prompts
                #
                while True:
                    if data[:2] == "> " or data[:2] == "+ ":
                        data = data[2:]
                    elif data[:1] == ">" or data[:1] == "+":
                        data = data[1:]
                    else:
                        break
                #
                # If we're not at the end of the buffer, copy
                # the line to the end
                #
                if not snek_repl_win.in_last_line():
                    snek_repl_win.eob()
                    snek_repl_win.insert_at_point(data)
                snek_repl_win.eob()
                snek_current_window.dispatch(ch)
                data += "\n"
                snek_device.command(data, intr="")
        else:
            snek_current_window.dispatch(ch)
        prev_exit = False
        prev_get = False


# Class to monitor the serial device for data and
# place in approprite buffer. Will be used as
# parameter to SnekDevice, and so it must expose
# 'cv' as a condition variable and 'receive' as a
# function to get data


class SnekMonitor:
    def __init__(self):
        global snek_lock
        self.cv = threading.Condition(snek_lock)
        self.ack = threading.Condition(snek_lock)
        self.dc4 = threading.Condition(snek_lock)
        self.getting_text = False

    # Synchronize with the remote device by sending ENQ and
    # waiting until an ACK is received

    def sync(self):
        global snek_lock
        # snek_debug('sync locking')
        with snek_lock:
            # snek_debug('sync locked')
            snek_device.serial.write(b"\x05")
            self.ack.wait(1)
            # snek_debug('sync done')

    def ping(self):
        # snek_debug('send PING')
        snek_device.serial.write(b"\x14\x0f\x03\x0e")
        ret = self.dc4.wait(0.05)
        return ret

    def wait(self, delay):
        return self.dc4.wait(delay)

    # Reading text to snek_edit_win instead of snek_repl_win

    def add_to(self, window, data):
        global snek_current_window, snek_repl_win
        follow = window == snek_repl_win and window.point == len(window.text)
        window.text += data
        if follow:
            window.point += len(data)
        window.repaint()
        if snek_current_window:
            snek_current_window.set_cursor()

    # Receive serial data

    def receive(self, data):
        global snek_edit_win, snek_repl_win, snek_lock
        data_edit = b""
        data_repl = b""
        for c in data:
            b = bytes((c,))

            # STX - start receiving eeprom contents
            if b == b"\x02":
                self.getting_text = True

            # ETX - stop receiving eeprom contents
            elif b == b"\x03":
                self.getting_text = False

            # ACK - device has seen ENQ
            elif b == b"\x06":
                # snek_debug('receive ACK')
                with snek_lock:
                    # snek_debug('notify ack')
                    self.ack.notify_all()

            # DC4 - device has seen DC4
            elif b == b"\x14":
                # snek_debug("receive PING")
                with snek_lock:
                    # snek_debug("notify PING")
                    self.dc4.notify_all()

            # Ignore all control chars other than newline
            elif b < b"\x20" and b != b"\n":
                pass

            # Otherwise, handle text
            else:
                if self.getting_text:
                    data_edit += b
                else:
                    data_repl += b

        # Append data to appropriate buffer

        with snek_lock:
            if data_edit:
                self.add_to(
                    snek_edit_win, str(data_edit, encoding="utf-8", errors="ignore")
                )
            if data_repl:
                self.add_to(
                    snek_repl_win, str(data_repl, encoding="utf-8", errors="ignore")
                )

    # Device has failed in reader or writer thread

    def failed(self, device):
        global snek_device, snek_lock

        # Need to acquire lock for other threads

        with snek_lock:
            if snek_device:
                snek_device.close()
                del snek_device
                snek_device = False
            # snek_debug('show error dialog')
            ErrorWin("Device %s failed" % device, inputthread=False)


def main():
    global snek_device, snek_edit_win, snek_monitor

    snek_monitor = SnekMonitor()

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "--list", action="store_true", help="List available serial devices"
    )
    arg_parser.add_argument("--port", help="Serial device")
    arg_parser.add_argument("file", nargs="*", help="Read file into edit window")
    args = arg_parser.parse_args()
    snek_device = False
    if args.list:
        for port in snek_list_ports():
            print("%-30.30s %s %s" % (port.name, port.description, port.hwid))
        sys.exit(0)
    if args.port:
        try:
            snek_device = SnekDevice(args.port, snek_monitor)
        except OSError as e:
            print(e.strerror, file=sys.stderr)
            sys.exit(1)
    text = ""
    if args.file:
        try:
            with open(args.file[0], "r") as myfile:
                text = myfile.read()
        except OSError as e:
            print("%s: %s", (e.filename, e.strerror), file=sys.stderr)
            sys.exit(1)
    try:
        screen_init(text)
        if snek_device:
            snek_device.start()
        run()
    finally:
        screen_fini()


main()



#!/usr/bin/python3
#
# Copyright © 2020 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

import random


def check(got, r):
    if got < 0 or r <= got:
        print("random value %d outside range %d" % (got, r))
        exit(1)


for r in (2, 163844, 100000, 8388607):
    for t in range(100):
        v = random.randrange(r)
        check(v, r)



#!/usr/bin/python3
#
# Copyright © 2019 Paulo Henrique Silva <ph.silva@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#


def fail(which):
    print("fail %s" % which)
    exit(1)


def check(a, b, which):
    if a > b or a < b:
        fail(which)


check(1_2_3, 123, "1_2_3 should be equal to 123")
check(10_000_000, 10000000, "10_000_000 should be equal to 10000000")



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

import time
import sys


def clear():
    print("\x1b[2J", end="")


def display_string(x, y, str):
    print("\x1b[%d;%dH%s" % (y, x, str), end="")


# A piece for position 'num' is num + 1 + num stars centered in a
# field of max * 2 + 1 characters with spaces on either side. This
# way, every piece is the same number of characters


def bar(n):
    b = ""
    for i in range(n):
        b += "█"
    return b


def make_string(n, c):
    s = ""
    for i in range(n):
        s += c
    return s


def make_piece(num, max):
    return (
        make_string(max - num - 1, " ")
        + bar(2 * num + 1)
        + make_string(max - num - 1, " ")
    )


def make_pieces(max):
    pieces = []
    for n in range(max):
        pieces += [make_piece(n, max)]
    return pieces


towers = []

# Left edge of the towers

left_x = 4

# Position of the bottom of the stacks, set at runtime

bottom_y = 0

# delay between moves in seconds

move_delay = 0.0250

# Distance between columns on the screen

col_spacing = 25


def tower_x(col):
    return col * col_spacing + left_x


# Position of the top of the tower on the screen Shorter
# towers start further down the screen


def tower_y(tower):
    return bottom_y - len(tower)


def clear_above(x, tower):
    display_string(x, tower_y(tower) - 1, "                   ")


# Display a piece from a tower


def display_one(x, y, tower):
    display_string(x, tower_y(tower) + y, tower[y])


# Display a tower


def display_tower(x, tower):
    for y in range(len(tower)):
        display_one(x, y, tower)


# Display all of the towers.


def display_towers(x, towers):
    for i in range(len(towers)):
        display_tower(x + i * col_spacing, towers[i])


def display_flush():
    display_string(1, bottom_y + 1, "")
    sys.stdout.flush()
    time.sleep(move_delay)
    return True


# Display all of the towers, then move the cursor out of the way and
# flush the output


def display_hanoi():
    display_towers(left_x, towers)
    display_flush()


# Reset towers to the starting state, with all of the pieces in the
# first tower and the other two empty


def reset_towers(len):
    global towers
    global bottom_y
    towers = [make_pieces(len), [], []]
    bottom_y = len + 3


def remove_top(tower):
    piece = towers[tower][0]
    new_tower = []
    for i in range(1, len(towers[tower])):
        new_tower += [towers[tower][i]]
    towers[tower] = new_tower
    return piece


def add_top(tower, piece):
    new_tower = [piece]
    for i in range(0, len(towers[tower])):
        new_tower += [towers[tower][i]]
    towers[tower] = new_tower


# Move a piece from the top of one tower to the top of another


def move_piece(frm, to):
    piece = remove_top(frm)
    add_top(to, piece)


# The implementation of the game


def _hanoi(n, frm, to, use):
    if n == 1:
        move_piece(frm, to)
        clear_above(tower_x(frm), towers[frm])
        display_one(tower_x(to), 0, towers[to])
        display_flush()
    else:
        _hanoi(n - 1, frm, use, to)
        _hanoi(1, frm, to, use)
        _hanoi(n - 1, use, to, frm)


# A pretty interface which
# resets the state of the game,
# clears the screen and runs
# the program


def hanoi(len):
    reset_towers(len)
    clear()
    display_hanoi()
    _hanoi(len, 0, 1, 2)


hanoi(8)



#!/usr/bin/python3
#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#


def fail(which):
    print("fail %s" % which)
    exit(1)


def check(a, b, which):
    if a > b or a < b:
        fail(which)


# 1.701412e38 is 0x7f000001. Make sure both of these values look like
# numbers

check(-1.701412e38, 2 * -8.50706e37, "-1.701412e38")
check(1.701412e38, 2 * 8.50706e37, " 1.701412e38")

# in 32-bit floats, these should be +/- inf

check(2 * -1.701412e38, 4 * -8.50706e37, "2 * -1.701412e38")
check(2 * 1.701412e38, 4 * 8.50706e37, "2 *  1.701412e38")



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

tuple = (1, 2, 3, 4, 5)
list = [6, 7, 8, 9, 10]
string = "abcde"

#
# Test fetching from ordered objects
#

for i in range(1, 5):
    assert tuple[i - 1] == i
    assert tuple[-i] == 6 - i

    assert list[i - 1] == 5 + i
    assert list[-i] == 11 - i

    assert string[i - 1] == chr(ord("a") + i - 1)
    assert string[-i] == chr(ord("f") - i)



servo = M4
motor = M1
front = A1
back = A2


def see(sensor):
    return read(sensor) > 0.1


def setservo(v):
    talkto(servo)
    if v < 0:
        setleft()
        v = -v
    else:
        setright()
    setpower(v)
    on()


def go(t, d):
    talkto(motor)
    s = front
    if d:
        setright()
        s = front
    else:
        setleft()
        s = back
    setpower(1)
    onfor(t)


def straight(t, d):
    setservo(0)
    go(t, d)


def left(t, d):
    setservo(1)
    go(t, d)


def right(t, d):
    setservo(-1)
    go(t, d)


def bounce():
    straight(random.randrange(10) / 10 + 1, True)
    right(random.randrange(10) / 10 + 1, False)
    left(random.randrange(10) / 10 + 1, True)
    straight(random.randrange(10) / 10 + 1, False)


while True:
    bounce()



#!/usr/bin/python3

import random
import time

random.seed(time.monotonic())

# "Shuffle" a list, but *wrong*.

narray = 16
a = [0] * narray
h = [0] * narray

for _ in range(100000):
    for i in range(narray):
        a[i] = i
    for i in range(narray):
        j = random.randrange(narray)
        tmp = a[j]
        a[j] = a[i]
        a[i] = tmp
    for i in range(narray):
        if a[i] == narray - 1:
            h[i] += 1
            break

print(h)



#!/usr/bin/env python3
#
# Copyright © 2019–2020 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

import argparse
import serial
import sys
import threading
import time

port = None


def sync():
    """Send a command to print a percent-sign, wait until it appears"""
    global port

    # wait for the device to get started
    time.sleep(1)
    port.write(b"\x0eprint(chr(37))\n")
    while True:
        data = port.read(1)
        if data == b"%":
            break
    while True:
        data = port.read(1)
        if data == b"\n":
            break


def run_test(send, expect):
    global port

    if not expect:
        print("test %r: " % (send,), end="")
        input()
    else:
        print("test %r" % (send,))

    port.write(send)

    if expect:
        line = port.read_until()
        if line != expect:
            print("test %r failed (got %r expected %r)" % (send, line, expect))
            return False
    else:
        line = port.read_until()

    return True


tests = (
    (b"1+1\n", b"> 2\r\n"),
    (b"talkto(NEOPIXEL); neopixel(((0.1,0.2,0.3),(0.3,0.2,0.1))); 1\n", b"> 1\r\n"),
    (b"talkto(NEOPIXEL); neopixel(((0,0,0),(0,0,0))); 1\n", None),
    (
        b"talkto(M1); on(); talkto(M2); on(); talkto(M3); on(); talkto(M4); on(); 1\n",
        b"> 1\r\n",
    ),
    (
        b"talkto(M1); off(); talkto(M2); off(); talkto(M3); off(); talkto(M4); off(); 1\n",
        None,
    ),
)


def main():
    global port, reader_thread
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--port", help="Serial device")
    args = arg_parser.parse_args()
    port = args.port
    if not port:
        port = "/dev/ttyACM0"
    port = serial.Serial(
        port=port,
        baudrate=115200,
        bytesize=serial.EIGHTBITS,
        parity=serial.PARITY_NONE,
        stopbits=serial.STOPBITS_ONE,
        xonxoff=True,
        rtscts=False,
        dsrdtr=False,
    )
    sync()
    failures = 0
    for test in tests:
        if not run_test(test[0], test[1]):
            failures = failures + 1
    sys.exit(failures)


main()



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

expected = 0
times = 0


def start(init, final, interval):
    global expected, times
    times = 0
    expected = (final - init) // interval
    if expected < 0:
        expected = 0


def iter():
    global times
    times += 1


def check(which):
    global times, expected
    if times != expected:
        print("fail %s. times %d expected %d" % (which, times, expected))
        exit(1)


start(0, 10, 1)
for i in range(10):
    iter()
check("range(10)")

start(0, 10, 1)
for i in range(0, 10):
    iter()
check("range(0,10)")

start(0, 10, 1)
for i in range(0, 10, 1):
    iter()
check("range(0,10,1)")

start(1, 10, 1)
for i in range(1, 10):
    iter()
check("range(1,10)")

start(1, 10, 1)
for i in range(1, 10, 1):
    iter()
check("range(1,10,1)")

start(0, 10, -1)
for i in range(0, 10, -1):
    iter()
check("range(0,10,-1)")

start(1, 10, -1)
for i in range(1, 10, -1):
    iter()
check("range(1,10,-1)")

start(10, 0, 1)
for i in range(10, 0):
    iter()
check("range(10,0)")

start(10, 0, 1)
for i in range(10, 0, 1):
    iter()
check("range(10,0,1)")

start(10, 0, -1)
for i in range(10, 0, -1):
    iter()
check("range(10,0,-1)")

start(10, 1, -1)
for i in range(10, 1, -1):
    iter()
check("range(10,1,-1)")



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#!/usr/bin/python3


def test():
    val = 5
    tries = 0
    while tries < 10:
        tries += 1
        for x in range(10):
            if x == val:
                val = 20
                break
        else:
            break
    return tries


if test() != 2:
    exit(1)



#
# Copyright © 2021 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

tests = {
    "[1, 2, 3]": [1, 2, 3],
    "('a', 'b', 'c')": ("a", "b", "c"),
    "{1: 'hello', 2: 'world'}": {1: "hello", 2: "world"},
    "hello world": "hello world",
    "1": "1",
    "0.5": 1 / 2,
}

for expect in tests:
    value = tests[expect]
    result = str(value)
    assert result == expect



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

led = D13


def blink():
    talkto(led)
    while True:
        on()
        time.sleep(0.5)
        off()
        time.sleep(0.5)


blink()



#
# Copyright © 2020 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#
# Test chaining comparison operators
#

assert 0 < 1 < 2
assert 0 < 1 <= 2
assert not (0 < 1 == 2)
assert 0 < 1 != 2
assert not (0 < 1 >= 2)
assert not (0 < 1 > 2)
assert 0 <= 1 < 2
assert 0 <= 1 <= 2
assert not (0 <= 1 == 2)
assert 0 <= 1 != 2
assert not (0 <= 1 >= 2)
assert not (0 <= 1 > 2)
assert not (0 == 1 < 2)
assert not (0 == 1 <= 2)
assert not (0 == 1 == 2)
assert not (0 == 1 != 2)
assert not (0 == 1 >= 2)
assert not (0 == 1 > 2)
assert 0 != 1 < 2
assert 0 != 1 <= 2
assert not (0 != 1 == 2)
assert 0 != 1 != 2
assert not (0 != 1 >= 2)
assert not (0 != 1 > 2)
assert not (0 >= 1 < 2)
assert not (0 >= 1 <= 2)
assert not (0 >= 1 == 2)
assert not (0 >= 1 != 2)
assert not (0 >= 1 >= 2)
assert not (0 >= 1 > 2)
assert not (0 > 1 < 2)
assert not (0 > 1 <= 2)
assert not (0 > 1 == 2)
assert not (0 > 1 != 2)
assert not (0 > 1 >= 2)
assert not (0 > 1 > 2)
assert "a" < "b" < "c"
assert "a" < "b" <= "c"
assert not ("a" < "b" == "c")
assert "a" < "b" != "c"
assert not ("a" < "b" >= "c")
assert not ("a" < "b" > "c")
assert "a" <= "b" < "c"
assert "a" <= "b" <= "c"
assert not ("a" <= "b" == "c")
assert "a" <= "b" != "c"
assert not ("a" <= "b" >= "c")
assert not ("a" <= "b" > "c")
assert not ("a" == "b" < "c")
assert not ("a" == "b" <= "c")
assert not ("a" == "b" == "c")
assert not ("a" == "b" != "c")
assert not ("a" == "b" >= "c")
assert not ("a" == "b" > "c")
assert "a" != "b" < "c"
assert "a" != "b" <= "c"
assert not ("a" != "b" == "c")
assert "a" != "b" != "c"
assert not ("a" != "b" >= "c")
assert not ("a" != "b" > "c")
assert not ("a" >= "b" < "c")
assert not ("a" >= "b" <= "c")
assert not ("a" >= "b" == "c")
assert not ("a" >= "b" != "c")
assert not ("a" >= "b" >= "c")
assert not ("a" >= "b" > "c")
assert not ("a" > "b" < "c")
assert not ("a" > "b" <= "c")
assert not ("a" > "b" == "c")
assert not ("a" > "b" != "c")
assert not ("a" > "b" >= "c")
assert not ("a" > "b" > "c")


def add(a, b):
    return a + b


assert add(1 == 2 != 0, 3) == 3
assert add(1 == 1 != 1, 4) == 4



#
# Copyright © 2020 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#

import math

# Make sure lists, tuples and parameter lists can have
# a trailing comma

assert (
    1,
    2,
) == (1, 2)
assert [
    1,
] == [1]
assert [
    1,
    2,
] == [1, 2]

assert {
    1: "hello",
} == {1: "hello"}
assert {
    1: "hello",
    2: "world",
} == {1: "hello", 2: "world"}

assert math.gcd(64, 12) == math.gcd(
    64,
    12,
)
assert math.sin(
    1,
) == math.sin(1)



# Bumper Car

power = 1


def MoveOneDir():
    talkto(M2)
    setpower(power)
    setleft()
    on()
    talkto(M3)
    setpower(power)
    setright()
    on()


def MoveOtherDir():
    talkto(M2)
    setpower(power)
    setright()
    on()
    talkto(M3)
    setpower(power)
    setleft()
    on()


def WaitUntilCloseOneDir():
    while read(A7) < 0.3:
        pass


def WaitUntilCloseOtherDir():
    while read(A1) < 0.3:
        pass


def WaitRandom():
    wait = 0.25 + random.randrange(10) / 10
    time.sleep(wait)


def Spin():
    talkto(M2)
    setpower(power)
    setright()
    on()
    talkto(M3)
    setpower(power)
    setright()
    on()


def SpinRandom():
    Spin()
    WaitRandom()


def BumperCar():
    while True:
        WaitUntilCloseOtherDir()

        # Back up a bit and spin around
        MoveOneDir()
        time.sleep(0.5)
        SpinRandom()

        MoveOneDir()
        WaitUntilCloseOneDir()

        # Back up a bit and spin around
        MoveOtherDir()
        time.sleep(0.5)
        SpinRandom()

        MoveOtherDir()


BumperCar()



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#
# Test math builtins
#

import math


def fail(which):
    print("fail %s" % which)
    exit(1)


#
# Number-theoretic and representation functions
#
if not math.ceil(0.5) == 1:
    fail("math.ceil(0.5) == 1")
if not math.ceil(-0.5) == 0:
    fail("math.ceil(-0.5) == 0")
if not math.copysign(1.1, -2) == -1.1:
    fail("math.copysign(1.1, -2) == -1.1")
if not math.copysign(-1.1, -2) == -1.1:
    fail("math.copysign(-1.1, -2) == -1.1")
if not math.copysign(1.1, 2) == 1.1:
    fail("math.copysign(1.1, 2) == 1.1")
if not math.copysign(-1.1, 2) == 1.1:
    fail("math.copysign(-1.1, 2) == 1.1")
if not math.fabs(-1.1) == 1.1:
    fail("math.fabs(-1.1) == 1.1")
if not math.fabs(1.1) == 1.1:
    fail("math.fabs(1.1) == 1.1")
if not math.factorial(9) == 362880:
    fail("math.factorial(9) == 362880")
if not math.floor(0.5) == 0:
    fail("math.floor(0.5) == 0")
if not math.floor(-0.5) == -1:
    fail("math.floor(-0.5) == -1")
if not math.frexp(12.5) == (0.78125, 4):
    fail("math.frexp(12.5) == (0.78125, 4)")
if not math.isclose(math.fsum([0.001] * 1000), 1):
    fail("math.isclose(math.fsum([0.01] * 100), 1)")
if not math.gcd(21, 14) == 7:
    fail("math.gcd(21,14) == 7")
big = 10.1
for i in range(40):
    big *= big
if not math.isclose(1, 1):
    fail("math.isclose(1,1)")
if not not math.isclose(1.1, 1):
    fail("not math.isclose(1.1,1)")
if not math.isclose(1.1, 1, rel_tol=0.2):
    fail("math.isclose(1.1,1,rel_tol=.2)")
if not math.isfinite(math.factorial(33)):
    fail("math.isfinite(math.factorial(33))")
if not not math.isfinite(big):
    fail("not math.isfinite(big)")
if not not math.isfinite(math.inf):
    fail("not math.isfinite(math.inf)")
if not math.isinf(math.inf):
    fail("math.isinf(math.inf)")
if not math.isinf(big):
    fail("math.isinf(math.inf)")
if not not math.isinf(math.factorial(33)):
    fail("not math.isinf(math.factorial(33))")
if not math.isnan(math.nan):
    fail("math.isnan(math.nan)")
if not math.isnan(math.inf - math.inf):
    fail("math.isnan(math.inf - math.inf)")
if not not math.isnan(big):
    fail("not math.isnan(math.inf)")
if not not math.isnan(1.0):
    fail("not math.isnan(1.0)")
if not math.ldexp(1, 2) == 4:
    fail("math.ldexp(1,2) == 4")
if not math.ldexp(1, -2) == 1 / 4:
    fail("math.ldexp(1,-2) == 1/4")
if not math.modf(1.25) == (0.25, 1.0):
    fail("math.modf(1.25) == (0.25, 1.0)")
if not math.remainder(5, 2) == 1:
    fail("math.remainder(5,2) == 1")
if not math.remainder(2.25, 0.5) == 0.25:
    fail("math.remainder(2.25,.5) == .25")
if not math.trunc(3.1) == 3:
    fail("math.trunc(3.1) == 3")
if not math.trunc(-3.1) == -3:
    fail("math.trunc(-3.1) == -3")
if not round(3.4) == 3:
    fail("round(3.4) == 3")
if not round(3.6) == 4:
    fail("round(3.6) == 4")
if not round(-3.4) == -3:
    fail("round(-3.4) == -3")
if not round(-3.6) == -4:
    fail("round(-3.6) == -4")
#
# Power and logarithmic functions
#
if not math.exp(0) == 1:
    fail("math.exp(0) == 1")
if not math.isclose(math.exp(1), math.e):
    fail("isclose(math.exp(1), math.e)")
if not math.isclose(math.exp(1) - 1, math.expm1(1)):
    fail("isclose(math.exp(1) - 1, math.expm1(1))")
if not math.isclose(math.log(1), 0):
    fail("math.isclose(math.log(1), 0)")
if not math.isclose(math.log(math.e), 1):
    fail("math.isclose(math.log(math.e), 1)")
if not math.isclose(math.log1p(1), math.log(2)):
    fail("math.isclose(math.log1p(1),math.log(2))")
if not math.isclose(math.log2(5.5), math.log(5.5) / math.log(2)):
    fail("math.isclose(math.log2(5.5),math.log(5.5)/math.log(2))")
if not math.isclose(math.log10(5.5), math.log(5.5) / math.log(10)):
    fail("math.isclose(math.log10(5.5),math.log(5.5)/math.log(10))")
if not math.isclose(math.pow(2.2, 3.3), math.exp(math.log(2.2) * 3.3)):
    fail("math.isclose(pow(2.2,3.3),math.exp(math.log(2.2) * 3.3))")
if not math.sqrt(4) == 2:
    fail("math.sqrt(4) == 2")
if not math.isclose(math.sqrt(2) * math.sqrt(2), 2):
    fail("isclose(math.sqrt(2) * math.sqrt(2), 2)")
#
# Trigonometric functions
#
if not math.isclose(math.acos(0), math.pi / 2):
    fail("math.isclose(math.acos(0), math.pi/2)")
if not math.isclose(math.acos(1), 0):
    fail("math.isclose(math.acos(1), 0)")
if not math.isclose(math.asin(0), 0):
    fail("math.isclose(math.asin(0), 0)")
if not math.isclose(math.asin(1), math.pi / 2):
    fail("math.isclose(math.asin(1), math.pi/2)")
if not math.isclose(math.atan(0), 0):
    fail("math.isclose(math.atan(0), 0)")
if not math.isclose(math.atan(1), math.pi / 4):
    fail("math.isclose(math.atan(1), math.pi/4)")
if not math.isclose(math.atan2(0, 1), 0):
    fail("math.isclose(math.atan2(0,1), 0)")
if not math.isclose(math.atan2(2, 2), math.pi / 4):
    fail("math.isclose(math.atan2(2,2), math.pi/4)")

if not math.isclose(math.cos(math.pi / 2), 0, abs_tol=1e-6):
    fail("math.isclose(math.cos(0), math.pi/2)")
if not math.isclose(math.cos(0), 1):
    fail("math.isclose(math.cos(1), 0)")
if not math.isclose(math.hypot(3, 4), 5):
    fail("math.isclose(math.hypot(3,4), 5)")
if not math.isclose(math.sin(0), 0):
    fail("math.isclose(math.sin(0), 0)")
if not math.isclose(math.sin(math.pi / 2), 1):
    fail("math.isclose(math.sin(1), math.pi/2)")
if not math.isclose(math.tan(0), 0):
    fail("math.isclose(math.tan(0), 0)")
if not math.isclose(math.tan(math.pi / 4), 1):
    fail("math.isclose(math.tan(1), math.pi/4)")
#
# Angular conversion
#
if not math.isclose(math.degrees(0), 0):
    fail("math.isclose(math.degrees(0),0)")
if not math.isclose(math.degrees(math.pi), 180):
    fail("math.isclose(math.degrees(math.pi),180)")
if not math.isclose(math.radians(0), 0):
    fail("math.isclose(math.radians(0),0)")
if not math.isclose(math.radians(180), math.pi):
    fail("math.isclose(math.radians(180), math.pi)")
#
# Hyperbolic functions
#
if not math.isclose(math.sinh(1), (math.e - 1 / math.e) / 2):
    fail("math.isclose(math.sinh(1), (math.e - 1/math.e) / 2)")
if not math.isclose(math.cosh(1), (math.e + 1 / math.e) / 2):
    fail("math.isclose(math.cosh(1), (math.e + 1/math.e) / 2)")
if not math.isclose(math.tanh(1), math.sinh(1) / math.cosh(1)):
    fail("math.isclose(math.tanh(1), math.sinh(1)/math.cosh(1))")
if not math.isclose(math.asinh(math.sinh(1)), 1):
    fail("math.isclose(math.asinh(math.sinh(1)), 1)")
if not math.isclose(math.acosh(math.cosh(1)), 1):
    fail("math.isclose(math.acosh(math.cosh(1)), 1)")
if not math.isclose(math.atanh(math.tanh(1)), 1):
    fail("math.isclose(math.atanh(math.tanh(1)), 1)")
#
# Special functions
#
if not math.isclose(math.erf(1), 0.8427007929497149):
    fail("math.isclose(math.erf(1), 0.8427007929497149)")
if not math.isclose(math.erfc(1), 1 - math.erf(1)):
    fail("math.isclose(math.erfc(1), 1-math.erf(1))")
if not math.isclose(math.gamma(10), math.factorial(9)):
    fail("math.isclose(math.gamma(10), math.factorial(9))")
if not math.isclose(math.lgamma(10), math.log(math.factorial(9))):
    fail("math.isclose(math.lgamma(10), math.log(math.factorial(9)))")
#
# Constants
#
if not math.isclose(math.pi * 2, math.tau):
    fail("math.isclose(math.pi * 2, math.tau)")



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

# This program is for the Crickit FeatherWing board

# Motor 1 is on the right side of the car
mr = (MOTOR1A, MOTOR1B)

# Motor 2 is on the left side of the car
ml = (MOTOR2A, MOTOR2B)

# SIGNAL1 is wired to a Phototransistor in the front of the car
pf = SIGNAL1

# SIGNAL2 is wired to a Phototransistor in the rear of the car
pr = SIGNAL2

# Pick some speeds
f_speed = 0.5
r_speed = 0.5
t_speed_f = 0.6
t_speed_s = 0.2

# Move straight forward
def forw():
    talkto(mr)
    setpower(f_speed)
    setright()
    on()
    talkto(ml)
    setpower(f_speed)
    setleft()
    on()


# Move straight back
def back():
    talkto(mr)
    setpower(r_speed)
    setleft()
    on()
    talkto(ml)
    setpower(r_speed)
    setright()
    on()


# Rotate left
def left():
    talkto(mr)
    setpower(t_speed_f)
    setright()
    on()
    talkto(ml)
    setpower(t_speed_f)
    setright()
    on()


# Rotate right
def right():
    talkto(mr)
    setpower(t_speed_f)
    setleft()
    on()
    talkto(ml)
    setpower(t_speed_f)
    setleft()
    on()


# Stop
def stop():
    talkto(ml)
    setpower(0)
    off()
    talkto(mr)
    setpower(0)
    off()


# Move forward until we approach something
def go_forw():
    forw()
    while read(pf) < 0.25:
        pass
    stop()


# Move backwards until we approach something
def go_back():
    back()
    while read(pr) < 0.25:
        pass
    stop()


# Hook the functions together in a simple
# way to demonstrate how they work
def bumper():
    while True:
        go_forw()
        back()
        time.sleep(0.5)
        stop()
        left()
        time.sleep(0.25)
        stop()
        go_forw()
        go_back()
        go_forw()


bumper()



#
# Copyright © 2019 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
i = 0
j = 0
while i < 10:
    i = i + 1
    if i == 5:
        break
    j = j + 1

assert j == 4



#
# Copyright © 2022 Keith Packard <keithp@keithp.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#
# Test string operations
#

abcde = "abcde"
vwxyz = "vwxyz"

# array operator
assert abcde[1] == "b"
assert abcde[-1] == "e"

# in operator
assert "bc" in abcde
assert "be" not in abcde
assert not "be" in abcde

# plus
assert abcde + vwxyz == "abcdevwxyz"
assert vwxyz + abcde == "vwxyzabcde"

# times
assert abcde * 3 == "abcdeabcdeabcde"
assert 3 * vwxyz == "vwxyzvwxyzvwxyz"
